/*
  Copyright 1996, 1997, 2002 Ronald S. Burkey.

  This file is part of Ron's Indexing Program (RIP).

  RIP is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  RIP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with RIP; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	rip.c 
  Target:	MS-DOS
  Compiler:	Borland Turbo C 2.0
  Purpose:	Text-file indexer/compressor
  Mods:		09/22/96 RSB	Began.
		09/29/96 RSB	Added compression.
		10/02/96 RSB	Added master index.
		10/03/96 RSB	Changed command-line arguments.
				Added RIPCOUNT as option "I".
				Allow certain short-word keepers.
				Numbers removed.
				Code modularized, I hope.
		10/06/96 RSB	Change command-line arguments again.
				Timestamps of files are automatically preserved.
				The input file is deleted if no error occurred.
				Context-sensitive huffman codes.
		10/08/96 RSB	Metric hashcodes.
		10/10/96 RSB	Rip-file indexes sorted by hashcode,block
				rather than block,hashcode.
		10/11/96 RSB	Incorporated, I hope into the 'S' function.
		10/12/96 RSB	Full use of hash-sorted indexes.
		10/13/96 RSB	Hashcode fully representing word identifiers,
				without interpolation.
		10/14/96 RSB	Began modularizing so that I could implement
				a browse function.
		-11/04/96 RSB	Speeded sort operation.
		11/05/96	RSB	Added bookmarks.
		11/10/96 RSB	Incorrect browser function at EOF fixed.
				END key added to browser.
				RIP.CFG added, including drive translation,
				default colors.  Completely updated descriptive
				text.
		11/11/96 RSB	Added option of extracting first n lines of
				all files.
		03/01/97 RSB	There was a bug in the 'S' option, in which
				any wimp words being on the search list would
				cause the search to fail.  For example, you
				could find "wizard oz" but not "wizard of oz",
				because "of" is a wimp.  This has been fixed.
				Max search-files increased from 256 to 2000.
				Multiple index segments introduced, via the
				RIP.LST file; drive translation disabled for
				now when RIP.LST is used, since the translation
				probably needs to be different for each
				segment.
		03/07/97 RSB   	Most code other than main program split off
				into RIP_2.C because of compiler memory.
				Summary option now includes the blank lines in
				the output, although not in the count.
				Added *2 and directory translation option to
				RIP.LST file, to allow running with index
				files on CD.  'S' option in browser now
				exact match, except case-insensitive.
		01/08/02 RSB	Reformatted in GNU style, and made some other
				minor cleanups in preparation for the first
				GPL'd release.		
  Design Goals
  ------------

  This is a text-indexing/retrieval/browsing system with the design goal
  of easily maintaining a database of all available English-language
  public domain etext, using PC-type computers, into the early 21st
  century and (with minor modification) into the indefinite future.  At the
  beginning of the project, this meant maintaining approximately a 1Gbyte
  2000-file database using a 100 MHz Pentium, and the point is that similar
  performance levels are supposed to be maintained into the future.  The
  database will grow larger, but computer speeds and mass storage will
  increase also, and the latter should compensate (hopefully more than
  compensate) for the former.  At the same time, acceptable though
  degraded performance levels should be available using CD-ROMs, 486 or
  386 clase machines, etc.

  The point in a project such as this is to openly accept certain
  performance limitations up front in order to avoid building in
  functional limitations that will render the system inoperative when the
  database size increases past a certain point.  For example, while it may
  be faster to perform various sorting and searching operations in memory
  rather than from disk, if this means that when the available memory in
  the machine is exhausted that the database stops working, this is no
  bargain.  Therefore, at every point the system has been crafted to work
  with databases of any size, with the main limitation being only the
  size of the hard disk.  Two other limitations have also been built into
  the system:  There is approximately a 256M limit on the size of the
  individual text files; and there is a total limit of 65536 total files.
  (Thus implying a limit on the database size of 16 Terabytes.)
  The former limit is probably all right for the indefinite future, but
  the 64K-files limit will probably have to be fixed around 2005.

  I looked for shareware tools for this project prior to working on it
  myself, and discovered that while many such programs existed, they
  were all highly unsatisfactory in practice.  I found many defects in
  these programs, and had to work very hard to avoid falling into the
  same traps they did.  Usually
  a program would suffer from all the defects mentioned below, although
  some suffered only a few.  I hope I have avoided all of them:

  1.  Very quick to add files when the database was small, but decreasing
  exponentially in speed as files new files were added.

  2.  Exhaustion of memory and eventual total failure as the database
  increased in size.

  3.  Failure of program or even system crash when indexing certain files.

  4.  No use of file compression.

  5.  Indexes of enormous size, so that the combined text and index might
  be twice the size of the original.

  6.  No integrated browser that I would care to use.  (I admit that this
  DOS-version browser isn't great, though it isn't bad, but I am working
  on a Windows version to compensate for that.)

  In fact, I never found a system that I would trust with even a 20M
  database, so in the end I needed to write my own.  Here are the basic
  engineering requirements for the task.  ("N" is the size of the
  database.)

  a.  Search time should be proportional to log N.

  b.  File preparation time (compression/indexing) should be proportional
  to N log N, and should be incremental so that it is not necessary to
  prepare the entire database in adding a new file.

  c.  Combined size of database (compressed files and index) should be
  some fixed proportion (but smaller) of the plain text.

  d.  RAM usage fixed, and not increasing with the size of the database.

  I have achieved these things.  On a 90 MHz Pentium, a 750MB/1500-file
  database can be totally prepared in 3-4 hours.  (I didn't time it.)
  A small number of files can be incrementally added in about 1/2 hour.
  The compressed database is about 75% of the size of the original.
  The program runs even on 512K machines, though 640K is helpful for
  file preparation.  Search time is a fraction of a second for negative
  single-word searches, though searches for phrases consisting of
  relatively common words can probably take up to a minute.

  SYSTEM OVERVIEW
  ---------------

  Text files are compressed/indexed using the "RIP C" command line option.
  The text is compressed using the context-sensitive Huffman codes
  described below, and then indexing information is added to the file.
  The original text file is deleted.  The compressed file has the extension
  "RIP".

  The text is processed in blocks of 4096 characters, although for
  indexing purposes the block size is extended somewhat so that the blocks
  overlap.  Pointer information concerning the blocks is stored so that the
  blocks can be individually accessed and decompressed (for viewing or
  searching) without the necessity of decompressing the entire file.  Each
  block is numbered with a 16-bit value.  Thus the 256M file limitation:
  4K*64K=256M.

  Indexing of a file occurs as follows.  First, "trash" words are discarded.
  These are supposed to be very common words (such as "and", "but", etc.)
  that occur so often it is better just to assume they are always present,
  rather than clogging up the index with them.  For execution speed, though,
  rather then discard the most common words we simply discard all words of
  3 characters or less.  (We keep all words containing Z, Q, J, or X as
  exceptions to this, so OZ or QUO would still be indexed.)  Next, each
  word is assigned a 32-bit hashcode, computed arithmetically from the
  ASCII representation.  Since there are only a few hundred thousand
  English words, and 4 billian 32-bit hashcodes, it is hoped that there
  is very little overlap and that the hashcodes are almost unique --
  one word, one hashcode.  Anyhow, a list is generated; for each unique
  hashcode appearing in the file is given the blocks in which the
  hashcode appears.  This list is in increasing order by hashcode, and
  secondarily in increasing order by block.  Finally, a "master index"
  is produced for the file giving simply all the hashcodes appearing in
  the file and the file offsets of their block lists.

  How would this index be used?  Suppose for example that you wanted to
  locate the word "Einstein".  You would compute the hashcode for
  "Einstein", and then examine the master index of the file.  The
  master index is a sorted list (by hashcode) of fixed-length records.
  A fast binary search would then determine instantly whether the
  word (or at least its almost-unique hashcode) appeared in the file.
  If the hashcode doesn't appear, the word is guaranteed not to appear.
  If the hashcode does appear, the word is probably present (with a very
  high degree of probability), but it can be determined fully as follows.
  Get the block list for the hashcode.  Decompress each of those blocks in
  turn and do a plain-text search.  Slower than a fully indexed system,
  of course, but still pretty fast.

  This is fast for an individual file, but slows down quite a lot if you
  have 1000+ files, particularly on CD-ROM, so there is an additional
  preparation step, "RIP I".  This produces a master index file for the
  database as a whole.  The master index for the database is almost
  exactly like the master index for an individual file, except that instead
  of giving the block numbers in which each hashcode appears, it gives
  the file numbers.  (There is a separate file listing all the text files,
  and the file numbers are determined by this.)  So the system can determine
  quickly all the files in which a given hashcode appears, and from there
  can determine quickly all the blocks in each each of these file.

  When the 64K-file limit has to be extended, it can be done entirely at
  the level of the master index.  The individual text files do not need
  conversion.

  COMPRESSION
  -----------
  The technique is simple.  I have pre-analyzed some text files, more
  or less at random from what was my reading list at the momemt, and have
  generated an English-language Huffman code on the assumption that this is
  a reasonable average case.  The compression achieved by this means wasn't
  amazingly good (only about a 45% reduction), but I found that I could
  improve it a lot (>55-60% reduction) if I had context-varying (but not
  adaptive) Huffman codes.  In other words, a different Huffman code is used
  for the character following an "A" than for one following a "B".  I
  actually recognize 29 different contexts: A-Z, "digit", "punctuation",
  and "other".  ("Punctuation" is .,?!:;)  It is easy to see why this works
  better.  For example, "Q" is almost always followed by "U".  Punctuation
  is almost always followed by whitespace.  Digits are seldom followed by
  alphabetic characters.  And so on.  This means that if you know that
  went before, you can encode what follows in less bits, on the average.
  A really interesting example is "E".  This has the shortest Huffman code
  on the average, since it is the most common letter.  And yet it rarely
  (in comparison) appears at the beginnings of words.  It is common only
  in the interiors of words.  The Huffman tables are given by ASCII files
  (RIP_A.DAT, ..., RIP_Z.DAT, RIPDIGIT.DAT, RIPPUNCT.DAT, RIPOTHER.DAT),
  so they can be replaced easily.  The *.DAT files are generally produced
  by machine, for example by the HUFFMANW.C type (n.b., probably huffman1.c
  was what was really meant here) programs.  However, the
  program automatically uses them to create a smaller binary file,
  RIP_ALLH.DAT, to speed up program loading.  To replace the Huffman tables
  it is therefore necessary to erase RIP_ALLH.DAT, replace the other
  RIP*.DAT files, and then to run RIP.  NOTE that this will render any
  already-compressed files permenently unreadable, since they will be
  decompressed with different Huffman tables than they were compressed with.

  Exact compressed file format:
  -----------------------------

  Bytes 0-3	"RIP4"
  Bytes 4-7	Number of uncomressed bytes, LSB first.
  Bytes 8-11	Block size, LSB first.
  Bytes 12-15	Pointer to master index data.
  Bytes 16-	Table of longs (LSB first) giving bit offsets of the
		blocks.  There can be a maximum of 4G bits, so the
		max filesize is 536M bytes.
  ... compression data ...
		(A bit stream of Huffman codes)
  ... index data ...
		(For each hashcode actually appearing in the file,
		a list of unsigneds, in increasing order, giving the
		block numbers in which the hashcode appears.  No
		counts or terminators are used.)
  ... master index data ...
		(First, long unsigned integer giving the total number
		of distinct hashcodes appearing in the file.  This
		includes a terminator hashcode of 0, which is otherwise
		an unused value.  This is followed by a table of
		records, with each record being a pair of longs.  The
		first long is a hashcode, and the second is the
		offset of this hashcode in the index data.  The
		hashcodes appear in increasing order both in the index
		data and in the master index data.  There is a
		Terminating record with hashcode 0 and offset pointing
		to the master index data itself.  This arrangement
		allows each calculation for each hashcode of how many
		blocks are used.)

  The total number of blocks isn't given, since it is the number of
  uncompressed bytes, divided by the blocksize and rounded up.

  RIP.CFG
  -------

  The RIP.CFG file controls the following configurable features of the
  program.

  1.  Drive translation. Since it may happen that the master index has been
  generated on a different (and slower) computer than is being used for
  browsing -- for example, from a CDROM -- it can easily happen that the
  *.RIP files are in the correct directories in RIP.FIL, but in the wrong
  drive letters, and that it is inconvenient to regenerate the master
  index.  We handle this by dynamically mapping the drive letters found
  in RIP.FIL.  Simply put lines in RIP.CFG of the form
	  letter->letter
  For example,
	  D->E
  This would cause every occurrence of D: in RIP.FIL to be replaced by E:.
  The substitution is done on-the-fly, and RIP.FIL itself is not actually
  modified.
  ... Later.  This has been totally removed.  See RIP.LST instead.

  2.  To set the background/foreground colors for the main menu, the
  browser command bar, and the browser text:
	  BCK_MENU=color
	  CLR_MENU=color
	  BCK_COMMAND=color
	  CLR_COMMAND=color
	  BCK_TEXT=color
	  CLR_TEXT=color
  The colors are the usual allowed IBM PC colors 0-15.

  RIP.LST
  -------

  It occurred to me that generating a single huge index file for the
  entire database is very inefficient for several reasons.  For one thing,
  when we surpass the 64K-file limit, the system will have to be
  re-written.  Second, since as the database grows and we make a new
  CD each time we pass a 650K limit, the databases for these CDs need only
  be generated once.  A more efficient scheme than generating a single
  huge index is to allow instead a master list of indices:  on this list we
  would have one entry for each finalized CD, and probably one for work in
  progress.  Therefore, we never have to generate an index larger than that
  needed for work in progress.  Also, this effectively increases the number
  of files allowed to 64K*max_indexes.

  Anyhow, that's what RIP.LST is:  a master list, with one line for each
  searchable index.  In its simplest form, each record simply is the base
  name of the index.  For example, if we copied all the indices (*.FIL and
  *.IND) into a single directory and called them VOL1.FIL/VOL1.IND,
  VOL2.FIL/VOL2.IND, ..., RIP.FIL/RIP.IND (the latter being for work in
  progress and the former for finalized CDs), we could simply have a
  RIP.LST like this:

	  RIP
	  VOL1
	  VOL2
	  .
	  .
	  .

  The databases would be searched in the order shown.

  One slight difficulty with this is the amount of disk space required by
  all these indices (30M+ for a typical CD index), so we might want to keep
  the CD indices on the CDs themselves.  If we do this, Some earlier
  versions of RIP used a master index named RIP.FIL/RIP2.IND (rather than
  RIP.FIL/RIP.IND).  To add the "2" to the IND-file name, we put a second
  field of "*" in the record:

	  RIP
	  E:RIP *

  Thus, the working database segment (that is, what will become the next
  CD, but which isn't big enough yet for that) has its index files
  (RIP.FIL/RIP.IND) on the hard disk, while the completed CD segment has
  its index files (RIP.FIL/RIP2.IND) on the CD.

  Another difficulty is that the drive designators in CD-based RIP.FILs
  are generally wrong, since the CDs are moved from computer to computer.
  The drive translation scheme formerly used (but not discontinued) in
  RIP.CFG couldn't handle this.  Furthermore, it has even happened that the
  base directory (as well as the drive designation) has been wrong on the
  CD, because the index was originally generated with the files on the
  hard disk, and then merely copied to the CD.  For example, the pathnames
  in the CD RIP.FIL might be D:\RIP\something, when what is really wanted
  is E:\something.  The "something" is correct, but the leading part of the
  pathname is wrong.  This can be handled by adding two more fields to
  the RIP.LST records:

	  RIP
	  E:RIP * D:\RIP E:

  This would simply replace D:\RIP by E: in the leading positions of the
  filenames found in E:RIP.FIL, but would let the pathnames in RIP.FIL
  unchanged.

  MASTER TOC
  ----------

  Helpful Hint:  In really large databases, when you are simply trying to
  find a specific book by title or author, a free-form search of the
  entire database is a hassle.  It would be more convenient to just search
  a title/author list.  Of course, there is no such thing.  However, the
  "RIP n" command can be used to generate a master list of the beginnings
  (say, the first 10 lines) of all the files, and if this master list itself
  is treated just like any old text file (i.e., compressed and indexed), and
  is the very first file in the first index, then the system will be able
  to complete title/author searches much quicker, since we are guaranteed
  that most titles and authors are in the first first file searched.  This
  allows the user to determine the desired filename, and from there to
  fetch it with the 'F' command rather than the 'S' command.

*/

#define ORIGINAL_RIP
/* Increase the size of the stack allocated by Turbo C. */
unsigned _stklen = 0x4000U;

#include "rip.h"

void
fetch_bookmarks (void)
{
  FILE *f;
  int i, j;
  unsigned long n;
  num_bookmarks = 0;
  f = fopen ("RIP.MRK", "r");
  if (f != NULL)
    {
      while (2 == fscanf (f, "%lu %s\n", &n, s))
	{
	  s[MAX_NAMELENGTH - 1] = 0;
	  strupr (s);
	  for (i = 0; i < num_bookmarks; i++)
	    if (!strcmp (s, bookmarks[i].filename))
	      break;
	  if (i == MAX_BOOKMARKS)
	    i = 0;
	  if (i < num_bookmarks)
	    {
	      num_bookmarks--;
	      for (j = i; j < num_bookmarks; j++)
		bookmarks[j] = bookmarks[j + 1];
	    }
	  bookmarks[num_bookmarks].offset = n;
	  strcpy (bookmarks[num_bookmarks].filename, s);
	  num_bookmarks++;
	}
      fclose (f);
    }
}

/* Filename translation.  This is used if, for instance, you have moved
   all or chunks of the database to different drives, but
   don't want to regenerate the master file list.  For example, if you
   have the database on CD, but the CDROM drive is at a different drive
   letter than on the computer that generated the index. */
static int xlation_index = 0;
char s1[256], s2[256], s3[256], s4[256];
void
filename_translate (char *filename)
{
  int n;
  if (xlation_index >= num_indices)
    return;
  strupr (filename);
  n = strlen (indices[xlation_index].src);
  if (!n)
    return;
  if (!strncmp (filename, indices[xlation_index].src, n))
    {
      strcpy (s1, filename + n);
      sprintf (filename, "%s%s", indices[xlation_index].dst, s1);
    }
}

/* Reads RIP.CFG. */
void
read_configuration (void)
{
  FILE *cfg;
  int i;
  char *ss, c1, c2;
  cfg = fopen ("RIP.CFG", "r");
  if (cfg == NULL)
    return;
  while (NULL != fgets (s, sizeof (s), cfg))
    {
      for (ss = s; isspace (*ss); ss++);
      if (*ss == ';' || *ss == 0)
	continue;
      if (2 == sscanf (ss, "%c->%c", &c1, &c2))
	{
	}
      else if (1 == sscanf (ss, "BCK_MENU=%d", &i))
	BCK_MENU = i;
      else if (1 == sscanf (ss, "BCK_COMMAND=%d", &i))
	BCK_COMMAND = i;
      else if (1 == sscanf (ss, "BCK_TEXT=%d", &i))
	BCK_TEXT = i;
      else if (1 == sscanf (ss, "CLR_MENU=%d", &i))
	CLR_MENU = i;
      else if (1 == sscanf (ss, "CLR_COMMAND=%d", &i))
	CLR_COMMAND = i;
      else if (1 == sscanf (ss, "CLR_TEXT=%d", &i))
	CLR_TEXT = i;
    }
  fclose (cfg);
}

#define MAX_HASHTHINGS 200
static HASHVALUE hashthings[MAX_HASHTHINGS];
HASHVALUE
hash (char *s)
{
  HASHVALUE hi;
  int i;
  for (i = 0, hi = 0; *s; s++)
    {
      if (isalpha (*s))
	{
	  hi += hashthings[i] * (toupper (*s) - '@');
	  i++;
	}
    }
  if (i && !hi)
    return (-1);
  else
    return (hi);
}

int get_onebyte (FILE * fin);
unsigned get_text (FILE * fin, unsigned long num_text, unsigned long loffset,
		   char *text_buf, unsigned num);
int browse (char *infile, FILE * fin,
	    unsigned long num_text, unsigned long offset);

void
copyright (void)
{
  cprintf ("RIP Text Indexing System, %s\r\n", __DATE__);
  cprintf ("(c)1996-2002 by Ron Burkey\r\n\n");
}

void
main (int argc, char *argv[])
{
  int err, retval;
  unsigned ii, i, j, k, l, min_ptr, last_line;
  unsigned long n, block, checksum, embedded_chk;
  char *ss, *sss;
  HASHVALUE hi;
  FILE *bookmark;

  read_configuration ();
  textcolor (CLR_MENU);
  textbackground (BCK_MENU);
  lowvideo ();

  for (hi = 1, i = 0; i < MAX_HASHTHINGS; i++, hi *= 29)
    hashthings[i] = hi;
  retval = 1;
  copyright ();

  /* check arguments */
  if (argc != argc)
    {
    usage_error:
      printf ("USAGE:\n");
      printf ("\tCompression:     RIP C <textfilelist\n");
      printf ("\t                 (Example:  dir /s /b *.txt | rip c  )\n");
      printf ("\tDecompression:   RIP D <ripfilelist\n");
      printf ("\t                 (Example:  dir /s /b *.rip | rip d  )\n");
      printf ("\tmaster Index:    RIP I <ripfilelist\n");
      printf ("\t                 (Example:  dir /s /b *.rip | rip i  )\n");
      printf ("\tinteractive:     RIP\n");
      printf ("\tHashcode check:  RIP H\n");
      printf ("\tsize report:     RIP %% <ripfilelist\n");
      printf ("\t                 (Example:  dir /s /b *.rip | rip %%  )\n");
      printf ("\tfile test:       RIP T <ripfilelist\n");
      printf ("\t                 (Example:  dir /s /b *.rip | rip T  )\n");
      printf ("\tn-line abstract: RIP n <ripfilelist\n");
      printf ("\t                 (Shows first n lines of every file)\n");
      printf ("The file RIP_ALLF.DAT must be available.\n");
      goto errxit;
    }

  {
    if (!strcmpi (argv[1], "H"))
      {
	for (;;)
	  {
	    printf ("Input: ");
	    gets (s);
	    hi = hash (s);
	    printf ("Hash=%ld %08lX\n", hi, hi);
	  }
      }
  }

  ftemp = fopen ("rip.tmp", "w+b");
  ftemp2 = fopen ("rip2.tmp", "w+b");
  if (ftemp == NULL)
    {
      printf ("Cannot create temporary file RIP.TMP or RIP2.TMP\n");
      goto errxit;
    }
  setvbuf (ftemp, NULL, _IOFBF, 8192);
  setvbuf (ftemp2, NULL, _IOFBF, 8192);

  /* Read the table of static Huffman codes. */
  fin = fopen ("rip_allf.dat", "rb");
  if (fin == NULL)
    {
      fin = fopen ("rip_allf.dat", "wb");
      for (i = 0; i < 29; i++)
	{
	  compressors[i] = farmalloc (sizeof (COMPRESSOR));
	  decompressors[i] = farmalloc (sizeof (DECOMPRESSOR));
	  if (compressors[i] == NULL || decompressors[i] == NULL)
	    {
	      printf ("Out of memory for Huffman tables.\n");
	      goto errxit;
	    }
	  if (i < 26)
	    {
	      sprintf (infile, "RIP_%c.DAT", 'A' + i);
	    }
	  else if (i == 26)
	    strcpy (infile, "RIPDIGIT.DAT");
	  else if (i == 27)
	    strcpy (infile, "RIPPUNCT.DAT");
	  else if (i == 28)
	    strcpy (infile, "RIPOTHER.DAT");
	  switch (read_huffman (infile, i))
	    {
	    case 1:
	      printf ("Huffman code table not found.\n");
	      goto errxit;
	    case 2:
	      printf ("Error in Huffman code file.\n");
	      goto errxit;
	    }
	  fwrite (compressors[i], sizeof (COMPRESSOR), 1, fin);
	  fwrite (decompressors[i], sizeof (DECOMPRESSOR), 1, fin);
	}
    }
  else
    {
      for (i = 0; i < 29; i++)
	{
	  compressors[i] = farmalloc (sizeof (COMPRESSOR));
	  decompressors[i] = farmalloc (sizeof (DECOMPRESSOR));
	  if (compressors[i] == NULL || decompressors[i] == NULL)
	    {
	      printf ("Out of memory for Huffman tables.\n");
	      goto errxit;
	    }
	  fread (compressors[i], sizeof (COMPRESSOR), 1, fin);
	  fread (decompressors[i], sizeof (DECOMPRESSOR), 1, fin);
	}
    }
  fclose (fin);
bypass:

  /* do the actual work */

  if (argc < 2)
    {
      FILE *files, *indexfile;
      unsigned fileno, fileptr;
      unsigned indexno;

      /* read index of indices */
      files = fopen ("RIP.LST", "r");
      if (files == NULL)
	{
	  strcpy (indices[0].name, "RIP");
	  num_indices = 1;
	}
      else
	{
	  while (NULL != fgets (s, sizeof (s), files))
	    {
	      for (ss = s; *ss; ss++)
		if (*ss == '\n')
		  *ss = 0;
	      *s1 = *s2 = *s3 = *s4 = 0;
	      i = sscanf (s, "%s %s %s %s", s1, s2, s3, s4);
	      if (!i)
		continue;
	      s1[MAX_INDEXNAME - 1] = 0;
	      strcpy (indices[num_indices].name, s1);
	      if (i > 1)
		{
		  if (*s2 == '*')
		    {
		      indices[num_indices].star2 = 1;
		      strcpy (s2, s3);
		      strcpy (s3, s4);
		      i--;
		    }
		  if (i > 1)
		    {
		      s2[MAX_INDEXTRAN - 1] = 0;
		      s3[MAX_INDEXTRAN - 1] = 0;
		      strupr (s2);
		      strupr (s3);
		      strcpy (indices[num_indices].src, s2);
		      strcpy (indices[num_indices].dst, s3);
		    }
		}
	      num_indices++;
	      if (num_indices >= MAX_INDICES)
		break;
	    }
	  fclose (files);
	}

      delete_file = 0;
      fileptr = 0;
      num_filesfound = 0;

    next_search:
      fetch_bookmarks ();

      clrscr ();
      copyright ();

      printf ("COMMAND MENU\n");
      printf ("\tB -- goto last Bookmark\n");
      printf ("\tF -- view File by filename\n");
      printf ("\tS -- word Search\n");
      printf ("\tR -- browse at random\n");
      printf ("\tX -- eXit\n");
      printf ("\t+ -- return to first file searched\n");
      printf ("\t- -- return to last file searched\n");
      for (i = 0; i < num_bookmarks; i++)
	printf ("\t%d -- %s bookmark\n", i, bookmarks[i].filename);
      printf ("CHOOSE: ");

      i = getch ();
    char_known:
      if (!i)
	{
	  getch ();
	  goto next_search;
	}
      i = toupper (i);
      if (i == '-')
	{
	  if (fileptr == num_filesfound && num_filesfound > 0)
	    fileptr--;
	  if (fileptr >= 0 && fileptr < num_filesfound)
	    goto search_entry;
	  goto next_search;
	}
      else if (i == '+')
	{
	  if (num_filesfound)
	    {
	      fileptr = 0;
	      goto search_entry;
	    }
	  goto next_search;
	}
      else if (i == 'X' || i == 27 || i == 'Q')
	{
	  clrscr ();
	  retval = 0;
	  bookmark = fopen ("RIP.MRK", "w");
	  for (i = 0; i < num_bookmarks; i++)
	    fprintf (bookmark,
		     "%lu %s\n", bookmarks[i].offset, bookmarks[i].filename);
	  fclose (bookmark);
	  goto errxit;
	}
      else if (i == 'F')
	{
	  putch (i);
	  cprintf ("\r\n\r\n");
	  n = 0;
	  printf ("Filename: ");
	  if (NULL == gets (s))
	    goto next_search;
	  if (1 == sscanf (s, "%s", infile))
	    {
	      strcat (infile, ".rip");
	      j = strlen (infile);
	      for (indexno = 0; indexno < num_indices; indexno++)
		{
		  xlation_index = indexno;
		  sprintf (outfile, "%s.FIL", indices[indexno].name);
		  files = fopen (outfile, "r");
		  if (files == NULL)
		    continue;
		  fseek (files, 0, SEEK_SET);
		  while (NULL != fgets (s, sizeof (s), files))
		    {
		      filename_translate (s);
		      if (!strncmpi (s, infile, j))
			goto f_success;
		      for (ss = s; *ss; ss++)
			if (*ss == ':' || *ss == '\\')
			  {
			    if (!strncmpi (ss + 1, infile, j))
			      {
				strcpy (ss + 1, infile);
				strcpy (infile, s);
			      f_success:
				fclose (files);
				goto browse_file;
			      }
			  }
		    }
		  fclose (files);
		}
	    }
	  goto next_search;
	}
      else if (i == 'B' || (i >= '0' && i <= '9'))
	{
	  if (i == 'B')
	    i = num_bookmarks - 1;
	  else
	    i -= '0';
	  if (i >= num_bookmarks || i < 0)
	    goto next_search;
	  strcpy (infile, bookmarks[i].filename);
	  n = bookmarks[i].offset;
	  goto browse_file;
	}
      else if (i == 'R')
	{
	  putch (i);
	  cprintf ("\r\n\r\n");
	  randomize ();
	  do
	    {
	      k = random (num_indices);
	      xlation_index = k;
	      sprintf (outfile, "%s.FIL", indices[k].name);
	      files = fopen (outfile, "r");
	    }
	  while (files == NULL);
	  fseek (files, 0, SEEK_END);
#pragma warn -sig
	  k = ftell (files) / 62;
#pragma warn +sig
	  k = random (k);
	  fseek (files, 62L * k, SEEK_SET);
	  fscanf (files, "%s", infile);
	  fclose (files);
	  filename_translate (infile);
	browse_file:
	  num_search = 2;
	  num_searchwords = 0;
	  if (open_file (infile))
	    goto next_search;
	  if (read_header ())
	    {
	      fclose (fin);
	      goto next_search;
	    }
	  if (i == 'R')
	    {
	      n = num_text / 100;
	      n *= random (100);
	    }
	  err = browse (infile, fin, num_text, n);
	  if (err == (A_S | 0x80))
	    {
	      fclose (fin);
	      for (sss = ss = infile; *sss; sss++)
		if (*sss == '\\' || *sss == ':')
		  ss = sss + 1;
	      mkdir ("c:\\unrip");
	      sprintf (outfile, "c:\\unrip\\%s", ss);
	      decompress_it (infile, outfile);
	    }
	  fclose (fin);
	  if (err == 'R')
	    {
	      i = err;
	      goto char_known;
	    }
	  else
	    goto next_search;
	}
      else if (i == 'S')
	{
	  putch (i);
	  cprintf ("\r\n\r\n");
	}
      else
	goto next_search;

      printf ("Search pattern: ");
      if (NULL == gets (s))
	goto next_search;
      err = sscanf (s, "%s %s %s %s %s %s %s %s %s %s %s %s %s %s",
		    &search[2].string, &search[3].string, &search[4].string,
		    &search[5].string, &search[6].string, &search[7].string,
		    &search[8].string, &search[9].string, &search[10].string,
		    &search[11].string, &search[12].string,
		    &search[13].string, &search[14].string,
		    &search[15].string);
      if (err == EOF || err == 0)
	goto next_search;
      num_search = err + 2;

      cprintf ("Opening catalogs ... ");

      for (i = 2; i < num_search; i++)
	{
	  if (iswimp (search[i].string))
	    {
	      num_search--;
	      for (j = i; j < num_search; j++)
		search[j] = search[j + 1];
	      i--;
	    }
	  else
	    {
	      search[i].hash = hash (search[i].string);
	    }
	  search[i].start_offset = 0;
	}

      num_filesfound = 0;
      for (indexno = 0; indexno < num_indices; indexno++)
	{
	  if (indices[indexno].star2)
	    sprintf (outfile, "%s2.IND", indices[indexno].name);
	  else
	    sprintf (outfile, "%s.IND", indices[indexno].name);
	  indexfile = fopen (outfile, "rb");
	  if (indexfile == NULL)
	    continue;
	  fseek (indexfile, 0, SEEK_SET);
	  fread (&n, sizeof (n), 1, indexfile);
	  fseek (indexfile, n, SEEK_SET);
	  for (fileno = findhash (indexfile); fileno != 0xffffU;
	       fileno = nexthash ())
	    {
	      if (num_filesfound >= MAX_FILESFOUND)
		{
		  printf ("Max file matches exceeded on search.\n");
		  break;
		}
	      else
		{
		  filesfound[num_filesfound].indexno = indexno;
		  filesfound[num_filesfound++].fileno = fileno;
		}
	    }
	  fclose (indexfile);
	}

      cprintf ("\r");
      clreol ();

      fileptr = 0;
    search_entry:
      textcolor (CLR_MENU);
      textbackground (BCK_MENU);
      lowvideo ();
      for (; fileptr < num_filesfound; fileptr++)
	{
	  cprintf ("\r%u/%u  \r", fileptr + 1, num_filesfound);
	  fileno = filesfound[fileptr].fileno;
	  indexno = filesfound[fileptr].indexno;

	  xlation_index = indexno;
	  sprintf (outfile, "%s.FIL", indices[indexno].name);
	  files = fopen (outfile, "r");
	  if (files == NULL)
	    goto no_matches;
	  fseek (files, 62L * fileno, SEEK_SET);
	  fscanf (files, "%s", infile);
	  fclose (files);
	  filename_translate (infile);
	  if (open_file (infile))
	    {
	      continue;
	    }
	  if (read_header ())
	    {
	      fclose (fin);
	      continue;
	    }
	  fseek (fin, index, SEEK_SET);

	  for (block = findhash (fin); block != 0xffffU; block = nexthash ())
	    {
	      unsigned long search_loc;

	      search_loc = ftell (fin);
#pragma warn -sig
	      get_text (fin, num_text, block * BLOCKSIZE, text_buf,
			OVERBLOCK);
	      text_buf[OVERBLOCK] = 0;
#pragma warn +sig
	      fseek (fin, search_loc, SEEK_SET);

	      /* Now check actual words as opposed to just hashes */
	      min_ptr = 0xffffU;
	      for (i = 2; i < num_search; i++)
		{
		  k = strlen (search[i].string);
#pragma warn -sig
		  j = OVERBLOCK - k - 1;
#pragma warn +sig
		  last_line = 0xffffU;
		  for (l = 0; l <= j;)
		    {
		      if (text_buf[l] == '\n')
			{
			  l++;
			  last_line = l;
			}
		      else if (isalpha (text_buf[l]))
			{
			  for (ii = 0;
			       !isspace (k = text_buf[l]) && k && l <= j; l++)
			    {
			      if (isalpha (k))
				{
				  s[ii++] = k;
				}
			    }
			  s[ii] = 0;
			  if (!strcmpi (search[i].string, s))
			    break;
			  if (!k)
			    {
			      l = j;
			      break;
			    }
			}
		      else
			l++;
		    }
		  if (l > j)
		    goto nextblock;
		  else if (search[i].hash)
		    {
		      if (last_line < min_ptr)
			min_ptr = last_line;
		    }
		}

	      if (min_ptr == 0xffffU)
		min_ptr = 0;
	      for (;;)
		{
		  err = toupper (browse (infile, fin, num_text,
					 block * BLOCKSIZE + min_ptr));
		  if (err == '.')
		    {
		      if (fileptr > 0)
			{
			  fileptr = -1;
			  goto nextfile;
			}
		    }
		  else if (err == '+')
		    goto nextfile;
		  else if (err == '-')
		    {
		      if (fileptr > 0)
			{
			  fileptr -= 2;
			  goto nextfile;
			}
		    }
		  else if (err == 27 || err == 'Q')
		    {
		      fclose (fin);
		      goto next_search;
		    }
		  else if (err == (A_S | 0x80))
		    {
		      fclose (fin);
		      for (sss = ss = infile; *sss; sss++)
			if (*sss == '\\' || *sss == ':')
			  ss = sss;
		      mkdir ("c:\\unrip");
		      sprintf (outfile, "c:\\unrip\\%s", ss);
		      decompress_it (infile, outfile);
		      fin = NULL;
		      fileptr--;
		      goto nextfile;
		    }
		}

	    nextblock:
	      k = 0;
	    }

	  /* no matches were found in this file */
	no_matches:
	  num_filesfound--;
	  for (i = fileptr; i < num_filesfound; i++)
	    filesfound[i] = filesfound[i + 1];
	  fileptr--;

	nextfile:
	  if (fin != NULL)
	    fclose (fin);
	  fin = NULL;
	  if (kbhit () && 27 == getch ())
	    break;

	}

      goto next_search;

    }
  else if (!strcmpi (argv[1], "%"))
    {
      unsigned long filecount;
      double compressed, uncompressed;

      filecount = 0;
      compressed = uncompressed = 0;
      while (NULL != gets (infile))
	{
	  printf ("File \"%s\", ", infile);
	  err = open_file (infile);
	  if (err)
	    {
	      printf ("Cannot open requested file.\n");
	      continue;
	    }
	  err = read_header ();
	  if (err)
	    {
	      fclose (fin);
	      printf ("Error in file format.\n");
	      continue;
	    }
	  fseek (fin, 0, SEEK_END);
	  n = ftell (fin);
	  fclose (fin);
	  printf ("%lu/%lu (%.1f%%)\n",
		  n, num_text,
		  ((signed long) (num_text - n)) * 100.0 / num_text);
	  filecount++;
	  compressed += n;
	  uncompressed += num_text;
	}
      printf ("%lu files, %.0f/%.0f bytes (%.1f%%)\n",
	      filecount, compressed, uncompressed,
	      100 * (uncompressed - compressed) / uncompressed);

    }
  else if (isdigit (*argv[1]))
    {

      while (NULL != gets (infile))
	{
	  void summarize (FILE *, unsigned, int);
	  printf ("\n<<File \"%s\">>\n\n", infile);
	  err = open_file (infile);
	  if (err)
	    {
	      printf ("Cannot open requested file.\n");
	      continue;
	    }
	  err = read_header ();
	  if (err)
	    {
	      fclose (fin);
	      printf ("Error in file format.\n");
	      continue;
	    }
	  summarize (fin, atoi (argv[1]), 1);
	  fclose (fin);
	}

    }
  else if (!strcmpi (argv[1], "D"))
    {

      delete_file = 1;
      while (NULL != gets (infile))
	{
	  strcpy (outfile, infile);
	  err = decompress_it (infile, outfile);
	  if (err)
	    cprintf ("Error=%04x\r\n", err);
	  if (fin != NULL)
	    fclose (fin);
	  if (fout != NULL)
	    fclose (fout);
	}

    }
  else if (!strcmpi (argv[1], "T"))
    {

      while (NULL != gets (infile))
	{
	  printf ("File \"%s\", ", infile);
	  err = open_file (infile);
	  if (err)
	    {
	      printf ("Cannot open requested file.\n");
	      continue;
	    }
	  err = read_header ();
	  if (err)
	    {
	      fclose (fin);
	      printf ("Error in file format.\n");
	      continue;
	    }
	  fseek (fin, -4, SEEK_END);
	  n = ftell (fin);
	  fread (&embedded_chk, sizeof (embedded_chk), 1, fin);
	  fseek (fin, 8, SEEK_SET);
	  checksum = 0;
	  for (block = 8; block < n; block++)
	    {
	      err = getc (fin);
	      if (err == EOF)
		break;
	      checksum += (unsigned) err;
	    }
	  if (checksum == embedded_chk)
	    printf ("okay\n");
	  else
	    printf ("error %lu,%lu\n", checksum, embedded_chk);
	  fclose (fin);
	}

    }
  else if (!strcmpi (argv[1], "C"))
    {

      delete_file = 1;
      while (NULL != gets (infile))
	{
	  strcpy (outfile, infile);
	  err = compress_it (infile, outfile);
	  if (err)
	    cprintf ("Error=%04x\r\n", err);
	  if (fin != NULL)
	    fclose (fin);
	  if (fout != NULL)
	    fclose (fout);
	}

    }
  else if (!strcmpi (argv[1], "I"))
    {

      err = index_it ();
      if (err)
	cprintf ("Error=%04x\r\n", err);
      fcloseall ();


    }
  else
    {

      fcloseall ();
      goto usage_error;
    }

errxit:
  fcloseall ();
  remove ("rip.tmp");
  remove ("rip2.tmp");
  exit (retval);
}


