/*
  Copyright 1996, 1997, 2002 Ronald S. Burkey.

  This file is part of Ron's Indexing Program (RIP).

  RIP is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  RIP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with RIP; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	rip.h 
  Purpose:	Header for rip.c and friends.
  Mods:		10/11/96 RSB	Started.
		03/07/97 RSB	RIP_TYPE added.
		01/08/02 RSB	Reformatted in GNU style and otherwise
				cleaned up in preparation for first
				GPL'd release.
		03/02/02 RSB	Headers rearranged as required by TurboC.
		03/03/02 RSB	Fixed field sizes in various structures,
				and added __attribute__ ((packed)) to them.
				
  This program was orginally targeted for MS-DOS, and compiled with
  Borland's Turbo C 2.x compiler.  It has been retargetd for GNU gcc, 
  using my TurboC porting library.  Still, there has been a lot of additional
  cleanup work.				
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <conio.h>
#include <dir.h>
#include <dos.h>
#include <alloc.h>
#include <io.h>
#include "fnkeys.h"

typedef long HASHVALUE;

#define ISVALID(x) (!isspace(x))
#define BLOCKSIZE 4096L

int read_huffman (char *filename, int c_index);
int iswimp (char *s);
HASHVALUE hash (char *s);
int rip (char *);
int read_header (void);
int create_file (char *fil);
int open_file (char *fil);
int compress_it (char *infile, char *outfile);
int decompress_it (char *infile, char *outfile);
int index_it (void);
int ispunctuation (char c);
int longsort4 (int32_t (*cmp) (const void *, const void *),
	       char *infile,
	       uint16_t recsize, uint32_t start, uint32_t end, char *outfile);
char *rfgets (char *, int, FILE *);

#define OVERBLOCK (BLOCKSIZE+256)	/* block size for indexing */
#define MINSTRING 4		/* min size of non-trash words */

/*
  The Huffman code tables used for compression and for decompression are
  in different formats.  For compression, for each possible character
  (0 to 255) we simply need to know how many bits are used, and what the
  bits are.  There are always less than 32 bits, so we can use a long
  to store the bits.  For decompression, since a variable number of bits
  is used, we need to make a binary tree.
*/
typedef struct
{
  uint16_t numbits __attribute__ ((packed));
  uint32_t bits __attribute__ ((packed));
}
COMPRESS;
typedef COMPRESS COMPRESSOR[256];

/* "zero" is the offset in the array storing the tree of the zero branch.
   if zero==0, then there is none.  Similarly, "one" is the offset of the
   one branch, and one=0 if there isn't one.  If both are zero, then
   "value" is the decoded byte.  Otherwise, "value" is undefined. */
typedef struct
{
  char value __attribute__ ((packed));
  uint16_t zero __attribute__ ((packed));
  uint16_t one __attribute__ ((packed));
}
DECOMPRESS;

/* MAXD is the maximum number of nodes in the tree.  It might seem as though
   256 is enough, but since each atomic node must have a non-atomic node
   pointing to it, 256 cannot be enough.  On the other hand, each non-atomic
   node must actually point to two other nodes (atomic or non-atomic), rather
   than just one, since otherwise the tree wouldn't be minimal.  In fact
   it is easy to see by induction that any minimal tree with N atomic nodes
   must have N-1 non-atomic nodes. */
#define MAXD 511
typedef DECOMPRESS DECOMPRESSOR[MAXD];
#define MAX_SEARCH 16

typedef struct
{
  uint16_t block __attribute__ ((packed));
  HASHVALUE hash __attribute__ ((packed));
}
HASH_RECORD;

typedef struct
{
  HASHVALUE hash __attribute__ ((packed));
  uint32_t offset __attribute__ ((packed));
}
NDX;

#define NDX_BUFSIZE 50
#define SEARCHSTRING_SIZE 129
typedef struct
{
  HASHVALUE hash __attribute__ ((packed));	/* =0 for wimps */
  uint32_t start_offset __attribute__ ((packed));
  uint32_t end_offset __attribute__ ((packed));
  char string[SEARCHSTRING_SIZE] __attribute__ ((packed));
}
SEARCH;

#define MAX_BOOKMARKS 10
#define MAX_NAMELENGTH 62
typedef struct
{
  uint32_t offset __attribute__ ((packed));
  char filename[MAX_NAMELENGTH] __attribute__ ((packed));
}
BOOKMARK;

#define MAX_SEARCHWORDS 10

#define MAX_FILESFOUND 2000
typedef struct
{
  char indexno __attribute__ ((packed));
  uint16_t fileno __attribute__ ((packed));
}
FOUND;

#define MAX_INDICES 100
#define MAX_INDEXNAME 21
#define MAX_INDEXTRAN 11
typedef struct
{
  char star2 __attribute__ ((packed));
  char name[MAX_INDEXNAME] __attribute__ ((packed));
  char src[MAX_INDEXTRAN] __attribute__ ((packed));
  char dst[MAX_INDEXTRAN] __attribute__ ((packed));
}
INDEXFILE;

/* ORIGINAL_RIP distinguishes between the source file rip.c and the other
   source files of the project. */
#ifdef ORIGINAL_RIP
#define RIP_TYPE
uint32_t DefaultTextMode = C80X50;
#else
#define RIP_TYPE extern
extern uint32_t DefaultTextMode;
#endif
RIP_TYPE DECOMPRESSOR *decompressors[29];
RIP_TYPE COMPRESSOR *compressors[29];
RIP_TYPE FILE *fout, *fin, *ftemp, *ftemp2;
RIP_TYPE char s[1000];
RIP_TYPE BOOKMARK bookmarks[MAX_BOOKMARKS];
RIP_TYPE uint16_t num_bookmarks;
RIP_TYPE uint32_t num_text, blocksize, vindex, blocks;
RIP_TYPE union
{
  char cb[0x2000];
  uint16_t ib[0x1000];
}
file_hashes;
#define file_block file_hashes.cb
RIP_TYPE uint16_t table_entries;
RIP_TYPE char text_buf[OVERBLOCK + 1];
RIP_TYPE int delete_file;
RIP_TYPE char outfile[129], infile[129];
RIP_TYPE SEARCH search[MAX_SEARCH];
RIP_TYPE uint16_t num_search;
RIP_TYPE char searchwords[MAX_SEARCHWORDS][SEARCHSTRING_SIZE];
RIP_TYPE int num_searchwords;
uint16_t findhash (FILE * f);
uint16_t nexthash (void);
RIP_TYPE FOUND filesfound[MAX_FILESFOUND];
RIP_TYPE uint16_t num_filesfound;
RIP_TYPE INDEXFILE indices[MAX_INDICES];
RIP_TYPE uint16_t num_indices;
#ifdef ORIGINAL_RIP
int CLR_COMMAND = LIGHTGREEN, CLR_TEXT = YELLOW, CLR_MENU = WHITE;
int BCK_COMMAND = BLACK, BCK_TEXT = BLACK, BCK_MENU = BLACK;
#else
extern int CLR_COMMAND, CLR_TEXT, CLR_MENU;
extern int BCK_COMMAND, BCK_TEXT, BCK_MENU;
#endif
